/*
 * Copyright (c) 2014 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution  and  use  in  source  and binary  forms,  with  or  without
 * modification, are permitted provided that the following conditions are met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice and this list of conditions.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice and  this list of  conditions in the  documentation and/or
 *      other materials provided with the distribution.
 *
 * THE SOFTWARE  IS PROVIDED "AS IS"  AND THE AUTHOR  DISCLAIMS ALL WARRANTIES
 * WITH  REGARD   TO  THIS  SOFTWARE  INCLUDING  ALL   IMPLIED  WARRANTIES  OF
 * MERCHANTABILITY AND  FITNESS.  IN NO EVENT  SHALL THE AUTHOR  BE LIABLE FOR
 * ANY  SPECIAL, DIRECT,  INDIRECT, OR  CONSEQUENTIAL DAMAGES  OR  ANY DAMAGES
 * WHATSOEVER  RESULTING FROM  LOSS OF  USE, DATA  OR PROFITS,  WHETHER  IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR  OTHER TORTIOUS ACTION, ARISING OUT OF OR
 * IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 *                                           Anthony Mallet on Fri May 23 2014
 */
#include "ac_tk3_flash.h"

#include <sys/types.h>
#include <sys/stat.h>

#include <err.h>
#include <fcntl.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>

#include "flash.h"




/* --- elf_init ------------------------------------------------------------ */

/* Open ELF file, perform some consistency checks, and initialize internal
 * state
 */
Elf *
elf_init(const char *input)
{
  Elf32_Ehdr *ehdr;
  const char *id;
  int elf_fd;
  Elf *elf;
  size_t s;

  if (elf_version(EV_CURRENT) == EV_NONE) {
    warnx("%s", elf_errmsg(-1));
    return NULL;
  }

  elf_fd = open(input, O_RDONLY);
  if (elf_fd < 0) { warn("%s", input); return NULL; }

  if ((elf = elf_begin(elf_fd, ELF_C_READ, NULL)) == NULL) {
    warnx("%s: %s", input, elf_errmsg(-1));
    return NULL;
  }
  if (elf_kind(elf) != ELF_K_ELF) {
    warnx("%s: wrong ELF format", input);
    return NULL;
  }

  /* architecture */
  id = elf_getident(elf, &s);
  if (!id) {
    warnx("%s: %s", input, elf_errmsg(-1));
    return NULL;
  }
  if (id[EI_CLASS] != ELFCLASS32 ||
      id[EI_DATA] != ELFDATA2LSB) {
    warnx("%s: wrong ELF format (expecting 32-bit LSB executable)", input);
    return NULL;
  }

  ehdr = elf32_getehdr(elf);
  if (!ehdr) {
    warnx("%s: cannot read ehdr: %s", input, elf_errmsg(-1));
    return NULL;
  }

  if (ehdr->e_type != ET_EXEC || ehdr->e_machine != EM_AVR) {
    warnx("%s: wrong architecture (expecting AVR executable file)", input);
    return NULL;
  }

  if (ehdr->e_phnum >= PN_XNUM) {
    warnx("%s: too many entries in the phdr", input);
    return NULL;
  }

  return elf;
}


/* --- elf_flash_data ------------------------------------------------------ */

char *
elf_loadable_data(Elf *elf, size_t *size, enum memk kind)
{
  char *data, *raw;
  Elf32_Ehdr *ehdr;
  Elf32_Phdr *phdr;
  int i, n;

  /* retrieve the program header table */
  ehdr = elf32_getehdr(elf);
  if (!ehdr) {
    warnx("cannot read extended header: %s", elf_errmsg(-1));
    return NULL;
  }

  phdr = elf32_getphdr(elf);
  if (!phdr) {
    warnx("cannot read program header: %s", elf_errmsg(-1));
    return NULL;
  }

  /* count sections */
  *size = 0;
  for (i = n = 0; i < ehdr->e_phnum; i++)
    if (phdr[i].p_type == PT_LOAD && phdr[i].p_filesz > 0) {
      if ((kind == FL_FLASH && (phdr[i].p_paddr>>16) == 0) ||
          (kind == FL_EEPROM && (phdr[i].p_paddr>>16) == 0x81))
        n++;
      else continue;

      if (*size < (phdr[i].p_paddr & 0xffff) + phdr[i].p_memsz)
        *size = (phdr[i].p_paddr & 0xffff) + phdr[i].p_memsz;

      printf("Found section %d, load 0x%x, size %d\n",
             n, phdr[i].p_paddr, phdr[i].p_memsz);
    }

  /* get data */
  raw = elf_rawfile(elf, NULL);
  data = calloc(*size, 1);
  if (!raw || !data) { warnx("out of memory"); return NULL; }

  for (i = n = 0; i < ehdr->e_phnum; i++)
    if (phdr[i].p_type == PT_LOAD && phdr[i].p_filesz > 0) {
      if ((kind == FL_FLASH && (phdr[i].p_paddr>>16) == 0) ||
          (kind == FL_EEPROM && (phdr[i].p_paddr>>16) == 0x81))
        memcpy(data + (phdr[i].p_paddr & 0xffff),
               raw + phdr[i].p_offset, phdr[i].p_filesz);
    }

  return data;
}
